import re
import unittest
from collections import OrderedDict

from conans.test.utils.tools import NO_SETTINGS_PACKAGE_ID, TestClient, TestServer


class RemoteChecksTest(unittest.TestCase):

    def test_binary_defines_remote(self):
        servers = OrderedDict([("server1", TestServer()),
                               ("server2", TestServer()),
                               ("server3", TestServer())])
        client = TestClient(servers=servers, inputs=3*["admin", "password"])
        conanfile = """from conans import ConanFile
class Pkg(ConanFile):
    pass"""
        client.save({"conanfile.py": conanfile})
        client.run("create . pkg/0.1@lasote/testing")
        client.run("upload Pkg* --all -r=server1 --confirm")
        client.run("upload Pkg* --all -r=server2 --confirm")

        # It takes the default remote
        client.run("remove * -f")

        # Exported recipe gets binary from default remote
        client.run("export . --name=pkg --version=0.1 --user=lasote --channel=testing")
        client.run("install --reference=pkg/0.1@lasote/testing")
        self.assertIn("pkg/0.1@lasote/testing from local cache - Cache", client.out)
        self.assertIn("pkg/0.1@lasote/testing:%s - Download" % NO_SETTINGS_PACKAGE_ID,
                      client.out)
        self.assertIn("pkg/0.1@lasote/testing: Retrieving package "
                      "%s from remote 'server1'" % NO_SETTINGS_PACKAGE_ID, client.out)

        # Explicit remote also defines the remote
        client.run("remove * -f")
        client.run("export . --name=pkg --version=0.1 --user=lasote --channel=testing")
        client.run("install --reference=pkg/0.1@lasote/testing -r=server2")
        self.assertIn("pkg/0.1@lasote/testing from local cache - Cache", client.out)
        self.assertIn("pkg/0.1@lasote/testing:%s - Download" % NO_SETTINGS_PACKAGE_ID,
                      client.out)
        self.assertIn("pkg/0.1@lasote/testing: Retrieving package "
                      "%s from remote 'server2'" % NO_SETTINGS_PACKAGE_ID, client.out)

        # Ordered search of binary works
        client.run("remove * -f")
        client.run("remove * -f -r=server1")
        client.run("export . --name=pkg --version=0.1 --user=lasote --channel=testing")
        client.run("install --reference=pkg/0.1@lasote/testing")
        self.assertIn("pkg/0.1@lasote/testing from local cache - Cache", client.out)
        self.assertIn("pkg/0.1@lasote/testing:%s - Download" % NO_SETTINGS_PACKAGE_ID, client.out)
        self.assertIn("pkg/0.1@lasote/testing: Retrieving package "
                      "%s from remote 'server2'" % NO_SETTINGS_PACKAGE_ID, client.out)

        # Download recipe and binary from the remote2 by iterating
        client.run("remove * -f")
        client.run("remove * -f -r=server1")
        client.run("install --reference=pkg/0.1@lasote/testing")
        self.assertIn("pkg/0.1@lasote/testing from 'server2' - Downloaded", client.out)
        self.assertIn("pkg/0.1@lasote/testing:%s - Download" % NO_SETTINGS_PACKAGE_ID, client.out)
        self.assertIn("pkg/0.1@lasote/testing: Retrieving package "
                      "%s from remote 'server2'" % NO_SETTINGS_PACKAGE_ID, client.out)

    def test_binaries_from_different_remotes(self):
        servers = OrderedDict()
        servers["server1"] = TestServer()
        servers["server2"] = TestServer()
        client = TestClient(servers=servers, inputs=2*["admin", "password"])
        conanfile = """from conans import ConanFile
class Pkg(ConanFile):
    options = {"opt": [1, 2, 3]}
"""
        client.save({"conanfile.py": conanfile})
        client.run("create . pkg/0.1@lasote/testing -o pkg:opt=1")
        client.run("upload pkg* --all -r=server1 --confirm")
        client.run("remove * -p -f")
        client.run("create . pkg/0.1@lasote/testing -o pkg:opt=2")
        package_id2 = re.search(r"pkg/0.1@lasote/testing:(\S+)", str(client.out)).group(1)
        client.run("upload Pkg* --all -r=server2 --confirm")
        client.run("remove * -p -f")

        # recipe is cached, takes binary from server2
        client.run("install --reference=pkg/0.1@lasote/testing -o pkg:opt=2 -r=server2")
        self.assertIn("pkg/0.1@lasote/testing from local cache - Cache", client.out)
        self.assertIn(f"pkg/0.1@lasote/testing:{package_id2} - Download", client.out)
        self.assertIn(f"pkg/0.1@lasote/testing: Retrieving package {package_id2} "
                      "from remote 'server2'", client.out)

        # Nothing to update
        client.run("install --reference=pkg/0.1@lasote/testing -o pkg:opt=2 -r=server2 -u")
        self.assertIn("pkg/0.1@lasote/testing from 'server2' - Cache", client.out)
        self.assertIn(f"pkg/0.1@lasote/testing:{package_id2} - Cache",
                      client.out)

        # Build missing
        client.run("install --reference=pkg/0.1@lasote/testing -o pkg:opt=3 -r=server2", assert_error=True)
        self.assertIn("ERROR: Missing prebuilt package for 'pkg/0.1@lasote/testing'", client.out)

        client.run("install --reference=pkg/0.1@lasote/testing -o pkg:opt=3", assert_error=True)
        self.assertIn("ERROR: Missing prebuilt package for 'pkg/0.1@lasote/testing'", client.out)
